/* qdb2txt: Disassemble games written with the CP/M version of the "Quill" 
           adventure game system; output to a PAW-style source script
    Copyright (C) 2023  John Elliott

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

typedef unsigned char zbyte;
typedef unsigned short zword;

#define BASE 0xF00

#define SEPARATOR ";-    -    -    -    -    -    -    " \
		  "-    -    -    -    -    -    -    \n"

static int verbose;

zbyte qdb[0xF100];
char *infile = NULL;
char *outfile = NULL;

FILE *fpout;

zword peekw(zword offset)
{
	return qdb[offset] + 256 * qdb[offset + 1];
}

void syntax(const char *av0)
{
	fprintf(stderr, "Syntax: %s file.qdb {output.qse}\n\n", av0);
	exit(1);
}

static int xpos;

static void dotext(zword offset)
{
	xpos = 0;
	while (1)
	{
		char c = ~qdb[offset];

		if (c < ' ' || c > 126)
		{
			if (c == 0x0d) { fprintf(fpout, "\n"); return; }
			else if (c == 0x0a) { fprintf(fpout, "\n"); xpos = 0; }
			else fprintf(fpout, "\\0x%02x", c);
		}
/* A line cannot start with a slash or a semicolon. If it's the 
 * case that the compiled adventure somehow contains one
 * anyway, escape it. */
		else if (xpos == 0 && (c == '/' || c == ';'))
		{ 
			fprintf(fpout, "\\%c", c); 
			++xpos; 
		}
/* That also means we need to escape backslashes */
		else if (c == '\\')
		{
			fprintf(fpout, "\\\\"); 
			++xpos; 
		}
		else	
		{ 
			fprintf(fpout, "%c", c);
			++xpos; 
		}
		offset++;
	}
}


void dump_text(zword count, zword table)
{	
	int n;
	zword addr;
	zword systab = peekw(table) - BASE;

	for (n = 0; n < qdb[count]; n++)
	{
		addr = peekw(systab + 2 * n);
		fprintf(fpout, "/%d\n", n);
		dotext(addr - BASE);
	}
}



void show_word(zbyte w)
{
	int m;
	zword table = peekw(0x15) - BASE;

	if (w == 0xFF) 
	{
		fprintf(fpout, "_   ");
		return;
	}
	while(qdb[table + 4] != 0)
	{
		if (qdb[table + 4] == w)
		{
			for (m = 0; m < 4; m++) 
			{
				fputc(~qdb[table+m], fpout);
			}
			return;
		}
		table += 5;
	}
/* Unknown word, render as (num) */
	fprintf(fpout, "(%d) ", w);
}


void printat(zbyte what, zword count, zword tab)
{
	char ch;
	zword addr;

	fprintf(fpout, "\t;");

	if (what >= qdb[count]) return; /* Out of range */

	addr = peekw(peekw(tab) - BASE + 2 * what) - BASE;
	do
	{
		ch = ~qdb[addr];
		if (isprint(ch)) fputc(ch, fpout);
		++addr;
	} while (ch != '\n' && ch != '\r');
}


void showobj(zword obj) { printat(obj, 3, 11); }
void showloc(zword loc) { printat(loc, 4, 13); }
void showmsg(zword msg) { printat(msg, 5, 15); }
void showsys(zword msg) { printat(msg, 6, 17); }



void show_condact(zword addr)
{
	int first = 1;

	while (1)
	{
		if (qdb[addr] == 0xFF)
		{
			fprintf(fpout, "\n");
			break;
		}
		if (!first) fprintf(fpout, "\t\t");
		first = 0;
		switch(qdb[addr])
		{
			case 0:	fprintf(fpout, "AT\t%d", qdb[addr + 1]); 
				if (verbose) showloc(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 1:	fprintf(fpout, "NOTAT\t%d", qdb[addr + 1]); 
				if (verbose) showloc(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 2:	fprintf(fpout, "ATGT\t%d", qdb[addr + 1]); 
				if (verbose) showloc(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 3:	fprintf(fpout, "ATLT\t%d", qdb[addr + 1]); 
				if (verbose) showloc(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 4:	fprintf(fpout, "PRESENT\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 5:	fprintf(fpout, "ABSENT\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 6:	fprintf(fpout, "WORN\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 7:	fprintf(fpout, "NOTWORN\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 8:	fprintf(fpout, "CARRIED\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 9:	fprintf(fpout, "NOTCARR\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 10: fprintf(fpout, "CHANCE\t%d\n", qdb[addr + 1]); 
				addr += 2;
				break;
			case 11: fprintf(fpout, "ZERO\t%d\n", qdb[addr + 1]); 
				addr += 2;
				break;
			case 12: fprintf(fpout, "NOTZERO\t%d\n", qdb[addr + 1]); 
				addr += 2;
				break;
			case 13: fprintf(fpout, "EQ\t%d %d\n", qdb[addr + 1], 
					qdb[addr + 2]); 
				addr += 3;
				break;
			case 14: fprintf(fpout, "GT\t%d %d\n", qdb[addr + 1], 
					qdb[addr + 2]); 
				addr += 3;
				break;
			case 15: fprintf(fpout, "LT\t%d %d\n", qdb[addr + 1], 
					qdb[addr + 2]); 
				addr += 3;
				break;
			case 16: fprintf(fpout, "WORD3\t%d", qdb[addr + 1]); 
				if (verbose) { fprintf(fpout, "\t;"); show_word(qdb[addr + 1]); }
				addr += 2;
				break;
			case 17: fprintf(fpout, "WORD4\t%d", qdb[addr + 1]); 
				if (verbose) { fprintf(fpout, "\t;"); show_word(qdb[addr + 1]); }
				addr += 2;
				break;
			case 18: fprintf(fpout, "INVEN\n"); 
				addr++;
				break;
			case 19: fprintf(fpout, "DESC\n"); 
				addr++;
				break;
			case 20: fprintf(fpout, "QUIT\n"); 
				addr++;
				break;
			case 21: fprintf(fpout, "END\n"); 
				addr++;
				break;
			case 22: fprintf(fpout, "DONE\n"); 
				addr++;
				break;
			case 23: fprintf(fpout, "OK\n"); 
				addr++;
				break;
			case 24: fprintf(fpout, "ANYKEY\n"); 
				addr++;
				break;
			case 25: fprintf(fpout, "SAVE\n"); 
				addr++;
				break;
			case 26: fprintf(fpout, "LOAD\n"); 
				addr++;
				break;
			case 27: fprintf(fpout, "TURNS\n"); 
				addr++;
				break;
			case 28: fprintf(fpout, "SCORE\n"); 
				addr++;
				break;
			case 29: fprintf(fpout, "CLS\n"); 
				addr++;
				break;
			case 30: fprintf(fpout, "DROPALL\n"); 
				addr++;
				break;
			case 31: fprintf(fpout, "AUTOG\n"); 
				addr++;
				break;
			case 32: fprintf(fpout, "AUTOD\n"); 
				addr++;
				break;
			case 33: fprintf(fpout, "AUTOW\n"); 
				addr++;
				break;
			case 34: fprintf(fpout, "AUTOR\n"); 
				addr++;
				break;
			case 35: fprintf(fpout, "PAUSE\t%d\n", qdb[addr + 1]); 
				addr += 2;
				break;
			case 36: fprintf(fpout, "BELL\n"); 
				addr++;
				break;
			case 37: fprintf(fpout, "GOTO\t%d", qdb[addr + 1]); 
				if (verbose) showloc(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 38: fprintf(fpout, "MESSAGE\t%d", qdb[addr + 1]); 
				if (verbose) showmsg(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 39: fprintf(fpout, "REMOVE\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 40: fprintf(fpout, "GET\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 41: fprintf(fpout, "DROP\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 42: fprintf(fpout, "WEAR\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 43: fprintf(fpout, "DESTROY\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 44: fprintf(fpout, "CREATE\t%d", qdb[addr + 1]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 45: fprintf(fpout, "SWAP\t%d %d", qdb[addr + 1], qdb[addr + 2]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 3;
				break;
			case 46: fprintf(fpout, "PLACE\t%d %d", qdb[addr + 1], qdb[addr + 2]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 3;
				break;
			case 47: fprintf(fpout, "SET\t%d\n", qdb[addr + 1]); 
				addr += 2;
				break;
			case 48: fprintf(fpout, "CLEAR\t%d\n", qdb[addr + 1]); 
				addr += 2;
				break;
			case 49: fprintf(fpout, "PLUS\t%d %d\n", qdb[addr + 1], qdb[addr + 2]); 
				addr += 3;
				break;
			case 50: fprintf(fpout, "MINUS\t%d %d\n", qdb[addr + 1], qdb[addr + 2]); 
				addr += 3;
				break;
			case 51: fprintf(fpout, "LET\t%d %d\n", qdb[addr + 1], qdb[addr + 2]); 
				addr += 3;
				break;
			case 52: fprintf(fpout, "NEWLINE\n");
				addr++;
				break;
			case 53: fprintf(fpout, "PRINT\t%d\n", qdb[addr + 1]); 
				addr += 2;
				break;
			case 54: fprintf(fpout, "SYSMESS\t%d", qdb[addr + 1]); 
				if (verbose) showsys(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 2;
				break;
			case 55: fprintf(fpout, "ISAT\t%d %d", qdb[addr + 1], qdb[addr + 2]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 3;
				break;
			case 56: fprintf(fpout, "COPYOF\t%d %d", qdb[addr + 1], qdb[addr + 2]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 3;
				break;
			case 57: fprintf(fpout, "COPYOO\t%d %d", qdb[addr + 1], qdb[addr + 2]); 
				if (verbose) showobj(qdb[addr + 1]);
				fputc('\n', fpout);
				addr += 3;
				break;
			case 58: fprintf(fpout, "COPYFO\t%d %d", qdb[addr + 1], qdb[addr + 2]); 
				if (verbose) showobj(qdb[addr + 2]);
				fputc('\n', fpout);
				addr += 3;
				break;
			case 59: fprintf(fpout, "COPYFF\t%d %d\n", qdb[addr + 1], qdb[addr + 2]); 
				addr += 3;
				break;
			case 60: fprintf(fpout, "ISDESC\n");
				addr++;
				break;
			case 61: fprintf(fpout, "EXTERN\t%d\n", qdb[addr + 1]);
				addr += 2;
				break;
			default: fprintf(fpout, "??%02x??\n", qdb[addr]);
				addr++;
				break;
		}
	}
}


void dump_logic(zword addr, int response)
{
	zword table = peekw(addr) - BASE;
	zword func;

	while (1)
	{
		if (qdb[table] == 0 && qdb[table + 1] == 0)
		{
			return;
		}
		if (response)
		{
			show_word(qdb[table]);
			fprintf(fpout, "\t");
			show_word(qdb[table+1]);
			fprintf(fpout, "\t");
			func = peekw(table + 2);
			show_condact(func - BASE);
			table += 4;
		}
		else
		{
			fprintf(fpout, "_\t_\t");
			func = peekw(table);
			show_condact(func - BASE);
			table += 2;
		}
	}

}


void dump_connections()
{
	int n;
	zword table = peekw(0x13) - BASE;
	zword conn;

	for (n = 0; n < qdb[4]; n++)
	{
		conn = peekw(table + 2 * n) - BASE;

		fprintf(fpout, "/%d", n);
		if (verbose) 
		{ 
			fprintf(fpout, "\t\t\t");	
			showloc(n);
		}
		fputc('\n', fpout);
		while(qdb[conn] != 0xFF)
		{
			fprintf(fpout, "\t");
			show_word(qdb[conn]);
			fprintf(fpout, "\t%d",qdb[conn+1]);			
			if (verbose) 
			{	 
				fprintf(fpout, "\t");	
				showloc(qdb[conn+1]);
			}
			fprintf(fpout, "\n");
			conn += 2;
		}
	}
}

void dump_obdefs()
{
	int n;
	zword places = peekw(0x17) - BASE;
	zword names = peekw(0x19) - BASE;
	zbyte loc;
	zbyte name;

	for (n = 0; n < qdb[3]; n++)
	{
		loc  = qdb[places + n];
		name = qdb[names + n];

		fprintf(fpout, "/%d\t", n);
		switch (loc)
		{
			case 252:	fprintf(fpout, "_"); break;	
			case 253:	fprintf(fpout, "WORN"); break;	
			case 254:	fprintf(fpout, "CARRIED"); break;	
			default:	fprintf(fpout, "%d", loc); break;
		}
		fputc('\t', fpout);
		show_word(name);
		if (verbose) 
		{
			fputc('\t', fpout);
			showobj(n);
		}
		fputc('\n', fpout);
	}
}


void dump_vocab()
{
	int  m;

	zword table = peekw(0x15) - BASE;

	while(qdb[table + 4] != 0)
	{
		for (m = 0; m < 4; m++) fputc(~qdb[table+m], fpout);
		fprintf(fpout, "\t%d\n", qdb[table + 4]);	
		table += 5;
	}	

}

int main(int argc, char **argv)
{
	int n;
	FILE *fp;

	for (n = 1; n < argc; n++)
	{
		if (!strcmp(argv[n], "-v") || !strcmp(argv[n], "-V")) verbose = 1;
		else if (infile == NULL) infile = argv[n];
		else if (outfile == NULL) outfile = argv[n];
		else syntax(argv[0]);
	}
	if (infile == NULL)
	{
		syntax(argv[0]);
	}
	fp = fopen(infile, "rb");
	if (!fp)
	{
		perror(infile);
		return 1;
	}	
	memset(qdb, 0, sizeof(qdb));
	fread(qdb, 1, sizeof(qdb), fp);
	fclose(fp);
	if (outfile == NULL)
	{
		outfile = "standard output";
		fpout = stdout;
	}
	else
	{
		fpout = fopen(outfile, "w");
		if (!fpout)
		{
			perror(outfile);
			exit(1);
		}
	}
	fprintf(fpout, "/CTL\t\t;Control Section\n");
	fprintf(fpout, "Q\t\t;Indicates Quill rather than PAW data\n");
	fprintf(fpout, "_\t\t;Null word character is an underline\n");
	fprintf(fpout, SEPARATOR);
	fprintf(fpout, "/VOC\t\t;Vocabulary\n");
	fprintf(fpout, ";\t\tMovements are numbered < 13\n");
	dump_vocab();
	fprintf(fpout, SEPARATOR);
	fprintf(fpout, "/STX\t\t;System messages\n");
	dump_text(6, 0x11);
	fprintf(fpout, SEPARATOR);
	fprintf(fpout, "/MTX\t\t;Messages\n");
	dump_text(5, 0x0F);
	fprintf(fpout, SEPARATOR);
	fprintf(fpout, "/OTX\t\t;Object descriptions\n");
	dump_text(3, 0x0B);
	fprintf(fpout, SEPARATOR);
	fprintf(fpout, "/LTX\t\t;Location descriptions\n");
	dump_text(4, 0x0D);
	fprintf(fpout, SEPARATOR);
	fprintf(fpout, "/CON\t\t;Connections\n");
	dump_connections();
	fprintf(fpout, SEPARATOR);
	fprintf(fpout, "/OBJ\t\t;Object definitions\n");
	fprintf(fpout, ";obj\tstarts\tword\n");
	fprintf(fpout, ";num\t  at\n");
	dump_obdefs();
	fprintf(fpout, SEPARATOR);
	fprintf(fpout, "/PRO\t0\t;Process 0 (ie response)\n");
	dump_logic(7, 1);
	fprintf(fpout, SEPARATOR);
	fprintf(fpout, "/PRO\t2\t;Process 2 (run every turn)\n");
	dump_logic(9, 0);

	return 0;
}

